from enum import Enum, auto
import skimage.restoration
import numpy as np
import rawpy
from tqdm import tqdm
from PIL import Image
from skimage import img_as_float

class Color(Enum):
    RED = auto()
    GREEN_RIGHT = auto()
    GREEN_BOTTOM = auto()
    BLUE = auto()

    def __str__(self):
        return self.name.lower()

# Among:
# - `wavelet`
# - `bilateral`
# - `tv_chambolle`
def denoise(imageNpArray, denoiserName):
    skImageRestorationDenoise = getattr(skimage.restoration, f'denoise_{denoiserName}')

    match denoiserName:
        case 'wavelet':
            imageDenoisedNpArray = skImageRestorationDenoise(imageNpArray, rescale_sigma=True)
        case 'bilateral':
            imageDenoisedNpArray = skImageRestorationDenoise(imageNpArray, sigma_color=0.05, sigma_spatial=15)
        case 'tv_chambolle':
            imageDenoisedNpArray = skImageRestorationDenoise(imageNpArray, weight=0.2)
    return imageDenoisedNpArray

class iterativeMean:
    mean = None
    numberOfElementsInMean = 0

    def add(self, element):
        if self.mean is None:
            self.mean = element
        else:
            self.mean = ((self.mean * self.numberOfElementsInMean) + element) / (self.numberOfElementsInMean + 1)
        self.numberOfElementsInMean += 1

RAW_IMAGE_FILE_EXTENSIONS = [
    'arw',
    'nef',
]

def getRawColorChannel(raw, color):
    colorDesc = raw.color_desc.decode('ascii')
    assert colorDesc == 'RGBG'
    assert np.array_equal(raw.raw_pattern, np.array([[0, 1], [3, 2]], dtype = np.uint8))
    # RG
    # GB
    rawImageVisible = raw.raw_image_visible.copy()

    redRawImageVisible = rawImageVisible[::2, ::2]
    greenRightRawImageVisible = rawImageVisible[::2, 1::2]

    greenBottomRawImageVisible = rawImageVisible[1::2, ::2]
    blueRawImageVisible = rawImageVisible[1::2, 1::2]

    match color:
        case Color.RED:
            imageNpArray = redRawImageVisible
        case Color.GREEN_RIGHT:
            imageNpArray = greenRightRawImageVisible
        case Color.GREEN_BOTTOM:
            imageNpArray = greenBottomRawImageVisible
        case Color.BLUE:
            imageNpArray = blueRawImageVisible
    return imageNpArray

def isARawImage(imageFilePath):
    return any([imageFilePath.lower().endswith(f'.{rawImageFileExtension}') for rawImageFileExtension in RAW_IMAGE_FILE_EXTENSIONS])

def getColorChannel(imageFilePath, color):
    if isARawImage(imageFilePath):
        with rawpy.imread(imageFilePath) as raw:
            imageNpArray = getRawColorChannel(raw, color)
    else:
        imagePil = Image.open(imageFilePath)
        imageNpArray = img_as_float(np.array(imagePil))
    return imageNpArray

def escapeFilePath(filePath):
    return filePath.replace('/', '_')


def getNewIndex(index, offset):
    newIndex = (index - offset) * 2 + offset
    return newIndex

def silentTqdm(data, desc = None):
    return data

def mergeSingleColorChannelImagesAccordingToBayerFilter(singleColorChannelImages, progress = False):
    tqdmVersion = tqdm if progress else silentTqdm
    for color in tqdmVersion(Color, 'Color'):
        colorImage = singleColorChannelImages[color]
        if color == list(Color)[0]:
            width, height = colorImage.shape
            multipleColorsImage = np.empty([dimension * 2 for dimension in colorImage.shape], dtype = np.float64)
        for y in tqdmVersion(range(height), 'Height'):
            for x in range(width):
                pixel = colorImage[x, y]
                '''
                Assume Bayer Filter:
                RG
                GB
                '''
                offsetX, offsetY = {
                    Color.RED: (0, 0),
                    Color.GREEN_RIGHT: (1, 0),
                    Color.GREEN_BOTTOM: (0, 1),
                    Color.BLUE: (1, 1),
                }[color]
                newX, newY = [getNewIndex(index, offset) for index, offset in [(x, offsetX), (y, offsetY)]]

                multipleColorsImage[newX, newY] = pixel
    return multipleColorsImage